# -*- coding: utf8 -*-
# Copyright (c) 2022 Nicholas de Jong

__title__ = "domain-email-records"
__author__ = "Nicholas de Jong <ndejong@threatpatrols.com>"
__version__ = "0.5.2"
__license__ = "BSD2"
